function [S,err] = kHerding_resample(X,T,sig)
% Draw samples from any empirical distributions (a set of data points) with
% Kernel Herding
% The kernel function is isotropic Gaussian with sigma=sig
%
%	Input:
%       X:      D x N matrix of a set of data points, each column is a data point
%       T:      total number of samples to generate
%       sig:    std of Gaussian kernel
%
%   Output:
%       S:      D x T matrix of samples, each column is a sample
%       err:    L2 error of the average feature when computeErr=1, zeros otherwise
%
%   k(s1,s2): kernel function

%   N: total number of training samples
%   n: index of training samples
%   t: index of samples

%% parameters
[D, N] = size(X);

doPCA = 1; % pre-processing: PCA

% sig = 10;
Nstart = min(1e5,N);    % # candidate staring points (chosen from X) for local maximization
startidx = sort(randsample(N,Nstart)); % indices of these starting points

% parameters for customized local search
epsf = 1e-8;    % precision of fs to stop customized local search
epss = 1e-8;    % precision of s to stop customized local search
itmax = 100;    % maximal nr. of iterations to stop customized local search
eta = 5e3;      % initial step size
tetamin = 1e-8; % minimal step size

computeErr = 1; % 1: report the error of average features, 0: not computing error
opti = 0;       % 0: choose a sample only from the set Xstart, no local maximization
                % 1: matlab built-in maximization function, slower but more reliable,
                % 2: customized maximization, faster

%% initialization
display('Initialization...');
if doPCA == 1
    [U,Lam] = svd(X - repmat(mean(X,2),1,size(X,2)),'econ');
    Lam = Lam / sqrt(size(X,2));
    X = Lam\U'*X;
end

S = zeros(D, T);
err = zeros(T,1);

% pre-compute some constants for speed
const1 = 1; %(sqrt(2*pi)*sig)^D;
sig2 = sig^2;
sX2 = sum(X.*X,1);
sS2 = zeros(1,T);

% matrices for state selection, fs = t*fsX - fsS
% fsX = zeros(1,Nstart);
% for i=1:Nstart
%     i
%     fsX(i) = mean(kernel(X(:,startidx(i)),X,sX2));
% end
fsX = fastsumkernel(X(:,startidx),X,sX2)/N; % fsX(i) = 1/N*\sum_n K(X(:,n), X(:,startidx(i)))
fsS = zeros(1,Nstart);                      % fsS(i) = \sum_{t=1}^{T-1} K(S_t, Xstart(:,i))

% values for error calculation
if computeErr == 1
    kXS = 0;
    kSS = 0;
    kXX = 0; %#ok<NASGU>
    if numel(startidx)==N
        kXX = mean(fsX); % sum_{n,n'} pX_n * pX_n' * K(X_n,X_n')
    else
        kXX = sum(fastsumkernel(X,X,sX2))/N^2;
%         for i=1:N-1
%             kXX = kXX + 2*sum(kernel(X(:,i),X(:,i+1:end),sX2(i+1:end))) + kernel(X(:,i),X(:,i),sX2(i));
%         end
%         kXX = kXX/N^2;
    end
end

itlinebuffer = zeros(1,20);
idxit = 1;
meanitline = 4;

%% main loop
display('Start drawing samples...');
for t=1:T

    % update fss
    if t>1
        fsS = fsS + kernel(S(:,t-1),X(:,startidx),sX2(startidx));
    end

    fs = t*fsX - fsS;

    % select best state
    [ignore, idx] = max(fs);
    S(:,t) = X(:,startidx(idx)); sS2(t) = S(:,t)'*S(:,t);

    fprintf('iteration: %d',t);

    % local maximization
    if opti == 1
        %built-in fminunc
        s0 = S(:,t);
        options = optimset('GradObj','on','MaxFunEvals',1e5,'MaxIter',1e5,'TolX',1e-6,'TolFun',1e-6,'Display','off','DerivativeCheck','off');
        s1 = fminunc(@fungradfs,s0,options);
        S(:,t) = s1; % store samples
        sS2(t) = S(:,t)'*S(:,t);

	elseif opti == 2
        % costumized local maximization with line search
        s0 = S(:,t);
        f0 = -1e30+1; %value of w^T f(s)
        s = s0; s0 = s0+1;
        f = -funfs(s);
        it = 0;
        while true

            it = it+1;

            % convergence condition
            if f-f0<epsf || norm(s-s0)<epss || it>itmax
                break;
            end
            f0 = f;
            s0 = s;

            % compute approximate gradient ds
            ds = -funds(s);

            % line search
            fbest = f;
            teta = eta;
            itline = 0;
            while true
                f = -funfs(s + teta * ds);
                if f-f0>0
                    if f>fbest
                        etabest = teta;
                        fbest = f;
                    else
                        s = s + etabest * ds;
                        f = fbest;
                        break;
                    end
                end
                teta = teta*(sqrt(5)-1)/2;
                itline = itline + 1;
                if teta < tetamin
                  s = s + etabest * ds;
                  f = fbest;
                  break
                end
            end
%             fprintf('\nline it:%d\tidxit:%d',itline,idxit);

            %%%% adjust initial step automatically
            itlinebuffer(mod(idxit-1,length(itlinebuffer))+1) = itline;
            idxit = idxit + 1;
            if mod(idxit, length(itlinebuffer)) == 0
                if mean(itlinebuffer)>meanitline+1
                    eta = eta*((sqrt(5)-1)/2);
                elseif mean(itlinebuffer)<min(meanitline-1,2)
                    eta = eta/((sqrt(5)-1)/2);
                end
            end
            %%%%


        end
        S(:,t) = s; % store samples
        sS2(t) = S(:,t)'*S(:,t);
        fprintf('\tit: %d',it);

    end

    fprintf('\n');

	%% compute error
	% update kXS, kSS
    if computeErr == 1
        kXS = kXS + sum(kernel(S(:,t),X,sX2));
        if t>1
            kSS = kSS + 2*sum(kernel(S(:,t),S(:,1:t-1),sS2(1:t-1)));
        end
        kSS = kSS + kernel(S(:,t),S(:,t),sS2(t));

        err(t) = kXX - 2*kXS/t/N + kSS/t^2;

        % draw plots, save data every a while
        if mod(t,100)==0
            if ~exist('hFig1', 'var')
              hFig1 = figure;
              xlabel('# Samples');
              ylabel('# \epsilon^{-1}');
            end
            figure(hFig1);
            plot(abs(err(1:t)).^(-1/2));
            drawnow;
        end
    end

end % end of main loop

if doPCA == 1
    X = U*Lam*X;
    S = U*Lam*S;
end


    function f=funfs(s)
        f = - (t*mean(kernel(s,X,sX2)) - sum(kernel(s,S(:,1:t-1),sS2(1:t-1))));
    end

    function [f,g]=fungradfs(s)
        f = funfs(s);
        g = funds(s);
    end

    function ds = funds(s)
        ds = -(t/N*dkernel(s,X,sX2) - dkernel(s,S(:,1:t-1),sS2(1:t-1)));
    end

    % first argument is Dx1, the second is Dxn
    function k=kernel(x,y,sy2)
        if size(y,2)==0
            k = 0; return;
        end

        k = 1/const1*exp(-1/2/sig2*(sum(x.*x) - 2*x'*y + sy2));
    end

    function dx=dkernel(x,y,sy2)
        if size(y,2)==0
            dx = 0; return;
        end

        term = exp(-1/2/sig2*(sum(x.*x) - 2*x'*y + sy2))';
        dx = -(x*sum(term)-y*term)/const1/sig2;
    end

    % fast computation when x is (D,Nx) and y is (D,Ny)
    function k=fastsumkernel(x,y,sy2)

        Nx_tmp=size(x,2); Ny_tmp=size(y,2);
        sx2 = sum(x.*x,1);

        k = zeros(1,Nx_tmp);
        step_tmp = floor(1e7/Ny_tmp);
        for is_tmp = 1:step_tmp:Nx_tmp
            ie_tmp = min(is_tmp+step_tmp-1,Nx_tmp);
            k(is_tmp:ie_tmp) = 1/const1*sum(exp(-1/2/sig2*(repmat(sx2(is_tmp:ie_tmp)',1,Ny_tmp)...
                                                        - 2*x(:,is_tmp:ie_tmp)'*y + repmat(sy2,ie_tmp-is_tmp+1,1))),2)';
        end
    end

%% original kernel function and its derivative for reference
%% should return the same value, easier to understand but slower
%     function k=kernel(x,y)
%         if size(y,2)>1
%             x = repmat(x,1,size(y,2));
%         elseif size(y,2)==0
%             k = 0; return;
%         end
%
%         k = 1/const1*exp(-1/2/sig2*sum((x-y).^2));
%     end
%
%     function dx=dkernel(x,y)
%         if size(y,2)>1
%             x = repmat(x,1,size(y,2));
%         elseif size(y,2)==0
%             dx = 0; return;
%         end
%
%         dx = -(x-y)/const1/sig2*exp(-1/2/sig2*sum((x-y).^2))';
%     end
%%

end

