function [S,err] = kHerding_sample_GM(obj,T,sig)
% Draw samples from the true distribution of a Gaussian Mixture model with Kernel Herding
%
%   Input:
%       obj:    object of a Guassian Mixture model, generated from gmdistribution
%       T:      total number of samples to generate
%       sig:    std of Gaussian kernel
%
%   Output:
%       S:      D x T matrix of samples, each column is a sample
%       err:    L2 error of the average feature when computeErr=1, zeros otherwise
%
%   k(s1,s2): kernel function

%% parameters
mu = obj.mu;        % means
[C, D] = size(mu);  % #components, #dimensions

Nstart = 1e4;   % # candidate staring points for local maximization

% parameters for customized local search
epsf = 1e-8;    % precision of fs to stop customized local search
epss = 1e-8;    % precision of s to stop customized local search
itmax = 100;    % maximal nr. of iterations to stop customized local search
eta = 5e3;      % initial step size
tetamin = 1e-8; % minimal step size

computeErr = 1; % 1: report the error of average features, 0: not computing error
opti = 2;       % 0: choose a sample only from the set Xstart, no local maximization
                % 1: matlab built-in maximization function, slower but more reliable,
                % 2: customized maximization, faster

%% initialization
display('Initialization...');
Xstart = random(obj,Nstart)';

S = zeros(D, T);
err = zeros(T,1);

% pre-compute some constants for speed
sig2 = sig^2;
const1 = (sqrt(2*pi)*sig)^D;
const2 = zeros(1,C);
invSig = zeros(D,D,C);
for c=1:C
    invSig(:,:,c) = (eye(D)-pinv(sig2*pinv(obj.Sigma(:,:,c))+eye(D)))/sig2;
    const2(c) = obj.PComponents(c)/sqrt(2*pi)^D*sqrt(det(invSig(:,:,c)));
end
sXstart2 = sum(Xstart.^2,1);
sS2 = zeros(1,T);

% matrices for state selection, fs = t*fsX - fsS

% fsX = zeros(1,Nstart);
% for i=1:Nstart
%     fsX(i) = kernelGM(Xstart(:,i));
% end
fsX = kernelGMbatch(Xstart);    % fsX(i) = \int_x p(x) * K(x, Xstart(:,i))
fsS = zeros(1,Nstart);          % fsS(i) = \sum_{t=1}^{T-1} K(S_t, Xstart(:,i))

% values for error calculation
if computeErr == 1
    kXS = 0;
    kSS = 0;
    kXX = 0;
    for c1=1:C
        for c2=1:C
            kXX = kXX+obj.PComponents(c1)*obj.PComponents(c2)/sqrt(2*pi)^D*sqrt(1/det(invSig(:,:,c1)+pinv(obj.Sigma(:,:,c2)))...
                    *det(invSig(:,:,c1))/det(obj.Sigma(:,:,c2)))...
                    *exp(-1/2*(-(invSig(:,:,c1)*mu(c1,:)'+obj.Sigma(:,:,c2)\mu(c2,:)')'*pinv(invSig(:,:,c1)+pinv(obj.Sigma(:,:,c2)))...
                               *(invSig(:,:,c1)*mu(c1,:)'+obj.Sigma(:,:,c2)\mu(c2,:)') + mu(c1,:)*invSig(:,:,c1)*mu(c1,:)'...
                               + mu(c2,:)*pinv(obj.Sigma(:,:,c2))*mu(c2,:)'));
        end
    end
end

itlinebuffer = zeros(1,20);
idxit = 1;
meanitline = 4;

%% main loop
display('Start drawing samples...');
for t=1:T

    % update fss
    if t>1
        fsS = fsS + kernel(S(:,t-1),Xstart,sXstart2);
    end

    fs = t*fsX - fsS;

    % select best starting state
    [ignore, idx] = max(fs);
    S(:,t) = Xstart(:,idx); sS2(t) = S(:,t)'*S(:,t);

    fprintf('iteration: %d',t);

    % local maximization
    if opti == 1
        %built-in fminunc
        s0 = S(:,t);
        options = optimset('GradObj','on','MaxFunEvals',1e5,'MaxIter',1e5,'TolX',1e-6,'TolFun',1e-6,'Display','off','DerivativeCheck','off');
        s1 = fminunc(@fungradfs,s0,options);
        S(:,t) = s1; % store samples
        sS2(t) = S(:,t)'*S(:,t);

	elseif opti == 2
        % costumized local maximization with line search
        s0 = S(:,t);
        f0 = -1e30+1; %value of w^T f(s)
        s = s0; s0 = s0+1;
        f = -funfs(s);
        it = 0;
        while true

            it = it+1;

            % convergence condition
            if f-f0<epsf || norm(s-s0)<epss || it>itmax
                break;
            end
            f0 = f;
            s0 = s;

            % compute approximate gradient ds
            ds = -funds(s);

            % line search
            fbest = f;
            teta = eta;
            itline = 0;
            while true
                f = -funfs(s + teta * ds);
                if f-f0>0
                    if f>fbest
                        etabest = teta;
                        fbest = f;
                    else
                        s = s + etabest * ds;
                        f = fbest;
                        break;
                    end
                end
                teta = teta*(sqrt(5)-1)/2;
                itline = itline + 1;
                if teta < tetamin
                  s = s + etabest * ds;
                  f = fbest;
                  break
                end
            end
%             fprintf('\nline it:%d\tidxit:%d',itline,idxit);

            %%%% adjust initial step automatically
            itlinebuffer(mod(idxit-1,length(itlinebuffer))+1) = itline;
            idxit = idxit + 1;
            if mod(idxit, length(itlinebuffer)) == 0
                if mean(itlinebuffer)>meanitline+1
                    eta = eta*((sqrt(5)-1)/2);
                elseif mean(itlinebuffer)<min(meanitline-1,2)
                    eta = eta/((sqrt(5)-1)/2);
                end
            end
            %%%%


        end
        S(:,t) = s; % store samples
        sS2(t) = S(:,t)'*S(:,t);
        fprintf('\tit: %d',it);

    end

    fprintf('\n');

    %% compute error
    % update kXS, kSS
    if computeErr == 1
        kXS = kXS + kernelGM(S(:,t));
        if t>1
            kSS = kSS + 2*sum(kernel(S(:,t),S(:,1:t-1),sS2(1:t-1)));
        end
        kSS = kSS + kernel(S(:,t),S(:,t),sS2(t));

        err(t) = kXX - 2*kXS/t + kSS/t^2;

        % draw plots, save data every a while
        if mod(t,100)==0
            if ~exist('hFig1', 'var')
              hFig1 = figure;
              xlabel('# Samples');
              ylabel('# \epsilon^{-1}');
              hFig2 = figure;
            end            
            figure(hFig1);
            plot(abs(err(1:t)).^(-1/2));
            figure(hFig2);
            plot(S(1,1:t),S(2,1:t),'x');
            drawnow;
        end
    end

end % end of main loop

    function [f,g]=fungradfs(s)
        f = funfs(s);
        g = funds(s);
    end

    function f=funfs(s)
        f = - (t*kernelGM(s) - sum(kernel(s,S(:,1:t-1),sS2(1:t-1))));
    end

    function ds = funds(s)
        ds = -(t*dkernelGM(s) - dkernel(s,S(:,1:t-1),sS2(1:t-1)));
    end

    % first argument is Dx1, the second is Dxn
    function k=kernel(x,y,sy2)
        if size(y,2)==0
            k = 0; return;
        end

        k = 1/const1*exp(-1/2/sig2*(sum(x.*x) - 2*x'*y + sy2));
    end

    function dx=dkernel(x,y,sy2)
        if size(y,2)==0
            dx = 0; return;
        end

        term = exp(-1/2/sig2*(sum(x.*x) - 2*x'*y + sy2))';
        dx = -(x*sum(term)-y*term)/const1/sig2;
    end

    function k=kernelGM(x)
        k = 0;
        for c_tmp=1:C
            k = k + const2(c_tmp)*exp(-1/2*(x-mu(c_tmp,:)')'*invSig(:,:,c_tmp)*(x-mu(c_tmp,:)'));
        end
    end

    function k=kernelGMbatch(x)
        k = zeros(1,size(x,2));
        for c_tmp=1:C
%             P_tmp = cholcov(invSig(:,:,c_tmp));
            [U_tmp, L_tmp] = svd(invSig(:,:,c_tmp));
            P_tmp = sqrt(L_tmp)*U_tmp';
            x0_tmp = zeros(D,size(x,2));
            for d_tmp=1:D
                x0_tmp(d_tmp,:)=x(d_tmp,:)-mu(c_tmp,d_tmp);
            end
            k = k + const2(c_tmp)*exp(-1/2*sum((P_tmp*x0_tmp).^2,1));
        end
    end

    function dx=dkernelGM(x)
        dx = zeros(D,1);
        for c_tmp=1:C
            dx = dx - const2(c_tmp)*exp(-1/2*((x-mu(c_tmp,:)')'*invSig(:,:,c_tmp)*(x-mu(c_tmp,:)')))...
                        *(invSig(:,:,c_tmp)*(x-mu(c_tmp,:)'));
        end
    end

%% original kernel function and its derivative for reference
%% should return the same value, easier to understand but slower
%     function k=kernel(x,y)
%         if size(y,2)>1
%             x = repmat(x,1,size(y,2));
%         elseif size(y,2)==0
%             k = 0; return;
%         end
%
%         k = 1/const1*exp(-1/2/sig2*sum((x-y).^2));
%     end
%
%     function dx=dkernel(x,y)
%         if size(y,2)>1
%             x = repmat(x,1,size(y,2));
%         elseif size(y,2)==0
%             dx = 0; return;
%         end
%
%         dx = -(x-y)/const1/sig2*exp(-1/2/sig2*sum((x-y).^2))';
%     end
%%

end

